"""Utils for cleaning and comparing email addresses."""

from __future__ import annotations

from mismo.lib.phone._core import PhoneMatchLevel as PhoneMatchLevel
from mismo.lib.phone._core import PhonesDimension as PhonesDimension
from mismo.lib.phone._core import clean_phone_number as clean_phone_number
from mismo.lib.phone._core import match_level as match_level
