from __future__ import annotations

from ._blocker import NameBlocker as NameBlocker
from ._clean import NameStruct as NameStruct
from ._clean import normalize_name as normalize_name
from ._compare import NameComparer as NameComparer
from ._compare import NameMatchLevel as NameMatchLevel
from ._compare import equal_forgiving_typo as equal_forgiving_typo
from ._dimension import NameDimension as NameDimension
from ._nicknames import are_aliases as are_aliases
from ._nicknames import is_nickname_for as is_nickname_for
