"""Dense (using array<numeric>) and sparse (using map<any_type, numeric>) vectors."""

from __future__ import annotations

from mismo.vector._vector import cosine_similarity as cosine_similarity
from mismo.vector._vector import dot as dot
from mismo.vector._vector import mul as mul
from mismo.vector._vector import norm as norm
from mismo.vector._vector import normalize as normalize
