"""Compare module.

This module contains functions and classes for comparing pairs of records.
"""

from __future__ import annotations

from mismo.compare._comparer import PComparer as PComparer
from mismo.compare._match_level import LevelComparer as LevelComparer
from mismo.compare._match_level import MatchLevel as MatchLevel
from mismo.compare._plot import compared_dashboard as compared_dashboard
